//
//  RapidMixThread/h
//  This implementation of rapidMix trains in a different thread and has a synchronized
//  update process with your main thread, which triggers a (sync) callback on doneTraining.
//  Allows you to train one model and keep using it while the next one trains, the new
//  model will be swapped in sync with the thread the update is called in once it's done.
//  Created by James on 19/11/2017.
//
//

#ifndef RapidMixThread_h
#define RapidMixThread_h

#include <algorithm>
#include <functional>
#include <unistd.h> // usleep
#include "ThreadedProcess.h"

template <typename T>
class RapidMixThread : public ThreadedProcess
{
public:
    T* model = nullptr; // The model to be accessed
    std::function<void ( void ) > doneTrainingCallback = nullptr; // CB on done training
    std::function<void ( T* modelToBeTrained ) > threadedTrainCallback = nullptr; // Callback entry in train thread to pass training data and train the new model before it gets swapped in to place, note that any adjustments to the passed trainingdata or other non threadsafe elements at this point will result in undefined behavior
    
    RapidMixThread ( void ) : ThreadedProcess ()
    {
        model = new T();
        training.store(false);
        doneTraining.store(false);
        startThread(); // Start training thread
    }
    
    ~RapidMixThread ( void )
    {
        stopThread();
        if (newModel)
            delete newModel;
        if (model)
            delete model;
    }
    
    void train ( void )
    {
        if (!training.load())
        {
            training.store(true);
        }
    }
    
    bool isTraining ( void ) const
    {
        return training.load();
    }
    
    void update ( void )
    { // To be called in your main loop
        if (doneTraining.load())
        {
            if (newModel)
            {
                std::swap(model, newModel);
            } else
                printf("New model was not created successfully...\n");
            
            if (doneTrainingCallback)
                doneTrainingCallback();
            doneTraining.store(false);
        }
    }
    
protected:
    void mainThreadCallback ( void )
    { // Training thread
        if (isTraining())
        {
            if (newModel)
                delete newModel;
            
            newModel = new T();
            if (threadedTrainCallback)
                threadedTrainCallback(newModel);
            
            training.store(false);
            doneTraining.store(true);
        }
        usleep(100000); // Sleep 100 ms
    }
    
    std::atomic<bool> training;
    std::atomic<bool> doneTraining;
    T* newModel = nullptr;
    
private:
};

#endif /* RapidMixThread_h */
